@@ -1,11 +1,15 @@
 pub mod commit_txs_scanner;
 pub mod entry;
 
+pub(crate) mod chunk;
 pub(crate) mod container;
 pub(crate) mod orphan;
 pub(crate) mod pending;
 pub(crate) mod proposed;
 
+#[cfg(test)]
+mod tests;
+
 pub use self::entry::TxEntry;
 
 /// Equal to MAX_BLOCK_BYTES / MAX_BLOCK_CYCLES, see ckb-chain-spec.
