@@ -3,8 +3,10 @@
 
 mod block_assembler;
 mod callback;
+mod chunk_process;
 mod component;
 pub mod error;
+mod persisted;
 pub mod pool;
 mod process;
 pub mod service;
