@@ -1,5 +1,6 @@
 use crate::block_assembler::{BlockAssembler, BlockTemplateCacheKey, TemplateCache};
 use crate::callback::Callbacks;
+use crate::component::chunk::DEFAULT_MAX_CHUNK_TRANSACTIONS;
 use crate::component::commit_txs_scanner::CommitTxsScanner;
 use crate::component::entry::TxEntry;
 use crate::component::orphan::Entry as OrphanEntry;
@@ -21,20 +22,26 @@ use ckb_store::ChainStore;
 use ckb_types::{
     core::{
         cell::{
-            get_related_dep_out_points, OverlayCellChecker, ResolvedTransaction,
+            get_related_dep_out_points, OverlayCellChecker, ResolveOptions, ResolvedTransaction,
             TransactionsChecker,
         },
-        BlockView, Capacity, Cycle, EpochExt, ScriptHashType, TransactionView, UncleBlockView,
-        Version,
+        hardfork::HardForkSwitch,
+        BlockView, Capacity, Cycle, EpochExt, HeaderView, ScriptHashType, TransactionView,
+        UncleBlockView, Version,
     },
-    packed::{Byte32, CellbaseWitness, OutPoint, ProposalShortId, Script},
+    packed::{Byte32, Bytes, CellbaseWitness, OutPoint, ProposalShortId, Script},
     prelude::*,
 };
 use ckb_util::LinkedHashSet;
-use ckb_verification::cache::CacheEntry;
+use ckb_verification::{
+    cache::{CacheEntry, Completed},
+    ContextualTransactionVerifier, ScriptVerifyResult, TimeRelativeTransactionVerifier,
+    TxVerifyEnv,
+};
 use faketime::unix_time_as_millis;
 use std::collections::HashSet;
 use std::collections::{HashMap, VecDeque};
+use std::convert::TryInto;
 use std::sync::atomic::Ordering;
 use std::sync::{atomic::AtomicU64, Arc};
 use std::time::Duration;
@@ -49,12 +56,28 @@ pub enum PlugTarget {
     Proposed,
 }
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum TxStatus {
     Fresh,
     Gap,
     Proposed,
 }
 
+pub(crate) enum ProcessResult {
+    Suspended,
+    Completed(Completed),
+}
+
+impl TxStatus {
+    fn with_env(self, header: &HeaderView) -> TxVerifyEnv {
+        match self {
+            TxStatus::Fresh => TxVerifyEnv::new_submit(header),
+            TxStatus::Gap => TxVerifyEnv::new_proposed(header, 0),
+            TxStatus::Proposed => TxVerifyEnv::new_proposed(header, 1),
+        }
+    }
+}
+
 impl TxPoolService {
     async fn get_block_template_cache(
         &self,
@@ -106,9 +129,24 @@ impl TxPoolService {
             .code_hash(config.code_hash.pack())
             .hash_type(hash_type.into())
             .build();
+        let message = if config.use_binary_version_as_message_prefix {
+            if config.message.is_empty() {
+                config.binary_version.as_bytes().pack()
+            } else {
+                [
+                    config.binary_version.as_bytes(),
+                    b" ",
+                    config.message.as_bytes(),
+                ]
+                .concat()
+                .pack()
+            }
+        } else {
+            config.message.as_bytes().pack()
+        };
         let cellbase_witness = CellbaseWitness::new_builder()
             .lock(cellbase_lock)
-            .message(config.message.as_bytes().pack())
+            .message(message)
             .build();
 
         BlockAssembler::build_cellbase(snapshot, snapshot.tip_header(), cellbase_witness)
@@ -143,6 +181,7 @@ impl TxPoolService {
         max_block_cycles: Cycle,
         cellbase: &TransactionView,
         uncles: &[UncleBlockView],
+        extension_opt: Option<Bytes>,
     ) -> Result<(HashSet<ProposalShortId>, Vec<TxEntry>, u64), AnyError> {
         let guard = self.tx_pool.read().await;
         let uncle_proposals = uncles
@@ -156,6 +195,7 @@ impl TxPoolService {
             cellbase.data(),
             uncles,
             &proposals,
+            extension_opt,
         )?;
 
         let (entries, size, cycles) =
@@ -187,6 +227,7 @@ impl TxPoolService {
         uncles: Vec<UncleBlockView>,
         bytes_limit: u64,
         version: Version,
+        extension: Option<Bytes>,
     ) -> Result<BlockTemplate, AnyError> {
         let consensus = snapshot.consensus();
         let tip_header = snapshot.tip_header();
@@ -199,23 +240,41 @@ impl TxPoolService {
         let dummy_cellbase_entry = TxEntry::dummy_resolve(cellbase.clone(), 0, Capacity::zero(), 0);
         let entries_iter = iter::once(dummy_cellbase_entry).chain(entries.into_iter());
 
+        let resolve_opts = {
+            let hardfork_switch = snapshot.consensus().hardfork_switch();
+            let epoch_number = current_epoch.number();
+            ResolveOptions::new().apply_current_features(hardfork_switch, epoch_number)
+        };
+
         let rtxs: Vec<_> = block_in_place(|| {
-            entries_iter.enumerate().filter_map(|(index, entry)| {
-                let overlay_cell_checker = OverlayCellChecker::new(&transactions_checker, snapshot);
-                if let Err(err) = entry.rtx.check(&mut seen_inputs, &overlay_cell_checker, snapshot) {
-                    error!(
-                        "resolve transactions when build block template, tip_number: {}, tip_hash: {}, error: {:?}",
-                        tip_header.number(), tip_hash, err
-                    );
-                    None
-                } else {
-                    if index != 0 {
-                        transactions_checker.insert(entry.transaction());
-                        template_txs.push(BlockAssembler::transform_tx(&entry, false, None))
+            entries_iter
+                .enumerate()
+                .filter_map(|(index, entry)| {
+                    let overlay_cell_checker =
+                        OverlayCellChecker::new(&transactions_checker, snapshot);
+                    if let Err(err) = entry.rtx.check(
+                        &mut seen_inputs,
+                        &overlay_cell_checker,
+                        snapshot,
+                        resolve_opts,
+                    ) {
+                        error!(
+                            "resolve transactions when build block template, \
+                             tip_number: {}, tip_hash: {}, error: {:?}",
+                            tip_header.number(),
+                            tip_hash,
+                            err
+                        );
+                        None
+                    } else {
+                        if index != 0 {
+                            transactions_checker.insert(entry.transaction());
+                            template_txs.push(BlockAssembler::transform_tx(&entry, false, None))
+                        }
+                        Some(entry.rtx)
                     }
-                    Some(entry.rtx)
-                }
-            }).collect()
+                })
+                .collect()
         });
 
         // Generate DAO fields here
@@ -245,6 +304,7 @@ impl TxPoolService {
             cellbase: BlockAssembler::transform_cellbase(&cellbase, None),
             work_id: work_id.into(),
             dao: dao.into(),
+            extension: extension.map(Into::into),
         })
     }
 
@@ -312,6 +372,8 @@ impl TxPoolService {
                 .prepare_block_template_uncles(&snapshot, &block_assembler)
                 .await;
 
+            let extension = None;
+
             let (proposals, entries, txs_updated_at) = self
                 .package_txs_for_block_template(
                     bytes_limit,
@@ -319,6 +381,7 @@ impl TxPoolService {
                     cycles_limit,
                     &cellbase,
                     &uncles,
+                    extension.clone(),
                 )
                 .await?;
 
@@ -334,6 +397,7 @@ impl TxPoolService {
                 uncles,
                 bytes_limit,
                 version,
+                extension,
             )?;
 
             self.update_block_template_cache(
@@ -349,7 +413,7 @@ impl TxPoolService {
         }
     }
 
-    async fn fetch_tx_verify_cache(&self, hash: &Byte32) -> Option<CacheEntry> {
+    pub(crate) async fn fetch_tx_verify_cache(&self, hash: &Byte32) -> Option<CacheEntry> {
         let guard = self.txs_verify_cache.read().await;
         guard.peek(hash).cloned()
     }
@@ -366,9 +430,9 @@ impl TxPoolService {
         .collect()
     }
 
-    async fn submit_entry(
+    pub(crate) async fn submit_entry(
         &self,
-        verified: CacheEntry,
+        verified: Completed,
         pre_resolve_tip: Byte32,
         entry: TxEntry,
         mut status: TxStatus,
@@ -387,7 +451,17 @@ impl TxPoolService {
         _submit_entry(&mut tx_pool, status, entry, &self.callbacks)
     }
 
-    async fn pre_check(&self, tx: TransactionView) -> Result<PreCheckedTx, Reject> {
+    pub(crate) async fn orphan_contains(&self, tx: &TransactionView) -> bool {
+        let orphan = self.orphan.read().await;
+        orphan.contains_key(&tx.proposal_short_id())
+    }
+
+    pub(crate) async fn chunk_contains(&self, tx: &TransactionView) -> bool {
+        let chunk = self.chunk.read().await;
+        chunk.contains_key(&tx.proposal_short_id())
+    }
+
+    pub(crate) async fn pre_check(&self, tx: TransactionView) -> Result<PreCheckedTx, Reject> {
         // Acquire read lock for cheap check
         let tx_pool = self.tx_pool.read().await;
         let snapshot = tx_pool.cloned_snapshot();
@@ -409,11 +483,45 @@ impl TxPoolService {
         Ok((tip_hash, snapshot, rtx, status, fee, tx_size))
     }
 
+    pub(crate) async fn enqueue_remote_chunk_tx(
+        &self,
+        tx: TransactionView,
+        remote: (Cycle, PeerIndex),
+    ) -> Result<(), Reject> {
+        // non contextual verify first
+        if let Err(reject) = non_contextual_verify(&self.consensus, &tx) {
+            if reject.is_malformed_tx() {
+                self.ban_malformed(remote.1, format!("reject {}", reject));
+            }
+            return Err(reject);
+        }
+        self.chunk.write().await.add_remote_tx(tx, remote);
+        Ok(())
+    }
+
+    pub(crate) async fn resumeble_process_tx(&self, tx: TransactionView) -> Result<(), Reject> {
+        let limit_cycles = self.tx_pool_config.max_tx_verify_cycles;
+        let ret = self._resumeble_process_tx(tx.clone(), limit_cycles).await;
+
+        match ret {
+            Ok(processed) => {
+                if let ProcessResult::Completed(completed) = processed {
+                    self.after_process(tx, None, &Ok(completed)).await;
+                }
+                Ok(())
+            }
+            Err(e) => {
+                self.after_process(tx, None, &Err(e.clone())).await;
+                Err(e)
+            }
+        }
+    }
+
     pub(crate) async fn process_tx(
         &self,
         tx: TransactionView,
         remote: Option<(Cycle, PeerIndex)>,
-    ) -> Result<CacheEntry, Reject> {
+    ) -> Result<Completed, Reject> {
         let ret = self._process_tx(tx.clone(), remote.map(|r| r.0)).await;
 
         self.after_process(tx, remote, &ret).await;
@@ -425,14 +533,23 @@ impl TxPoolService {
         &self,
         tx: TransactionView,
         remote: Option<(Cycle, PeerIndex)>,
-        ret: &Result<CacheEntry, Reject>,
+        ret: &Result<Completed, Reject>,
     ) {
         let tx_hash = tx.hash();
+        // The network protocol is switched after tx-pool confirms the cache,
+        // there will be no problem with the current state as the choice of the broadcast protocol.
+        let with_vm_2021 = {
+            let epoch = self.snapshot().tip_header().epoch().number();
+            self.consensus
+                .hardfork_switch
+                .is_vm_version_1_and_syscalls_2_enabled(epoch)
+        };
+
         match remote {
             Some((declared_cycle, peer)) => match ret {
                 Ok(verified) => {
                     if declared_cycle == verified.cycles {
-                        self.broadcast_tx(Some(peer), tx_hash);
+                        self.broadcast_tx(Some(peer), tx_hash, with_vm_2021);
                         self.process_orphan_tx(&tx).await;
                     } else {
                         warn!(
@@ -450,7 +567,7 @@ impl TxPoolService {
                 }
                 Err(reject) => {
                     if is_missing_input(&reject) && self.all_inputs_is_unknown(&tx) {
-                        self.add_orphan(tx, peer).await;
+                        self.add_orphan(tx, peer, declared_cycle).await;
                     } else if reject.is_malformed_tx() {
                         self.ban_malformed(peer, format!("reject {}", reject));
                     }
@@ -458,13 +575,13 @@ impl TxPoolService {
             },
             None => {
                 match ret {
-                    Ok(_verified) => {
-                        self.broadcast_tx(None, tx_hash);
+                    Ok(_) => {
+                        self.broadcast_tx(None, tx_hash, with_vm_2021);
                         self.process_orphan_tx(&tx).await;
                     }
                     Err(Reject::Duplicated(_)) => {
                         // re-broadcast tx when it's duplicated and submitted through local rpc
-                        self.broadcast_tx(None, tx_hash);
+                        self.broadcast_tx(None, tx_hash, with_vm_2021);
                     }
                     Err(_err) => {
                         // ignore
@@ -474,8 +591,16 @@ impl TxPoolService {
         }
     }
 
-    pub(crate) async fn add_orphan(&self, tx: TransactionView, peer: PeerIndex) {
-        self.orphan.write().await.add_orphan_tx(tx, peer)
+    pub(crate) async fn add_orphan(
+        &self,
+        tx: TransactionView,
+        peer: PeerIndex,
+        declared_cycle: Cycle,
+    ) {
+        self.orphan
+            .write()
+            .await
+            .add_orphan_tx(tx, peer, declared_cycle)
     }
 
     pub(crate) async fn find_orphan_by_previous(
@@ -499,20 +624,35 @@ impl TxPoolService {
 
         while let Some(previous) = orphan_queue.pop_front() {
             if let Some(orphan) = self.find_orphan_by_previous(&previous).await {
-                match self._process_tx(orphan.tx.clone(), None).await {
-                    Ok(_) => {
-                        self.remove_orphan_tx(&orphan.tx.proposal_short_id()).await;
-                        self.broadcast_tx(Some(orphan.peer), orphan.tx.hash());
-                        orphan_queue.push_back(orphan.tx);
-                    }
-                    Err(reject) => {
-                        if !is_missing_input(&reject) {
+                if orphan.cycle > self.tx_pool_config.max_tx_verify_cycles {
+                    self.remove_orphan_tx(&orphan.tx.proposal_short_id()).await;
+                    orphan_queue.push_back(orphan.tx.clone());
+                    self.chunk
+                        .write()
+                        .await
+                        .add_remote_tx(orphan.tx, (orphan.cycle, orphan.peer));
+                } else {
+                    match self._process_tx(orphan.tx.clone(), None).await {
+                        Ok(_) => {
+                            let with_vm_2021 = {
+                                let epoch = self.snapshot().tip_header().epoch().number();
+                                self.consensus
+                                    .hardfork_switch
+                                    .is_vm_version_1_and_syscalls_2_enabled(epoch)
+                            };
                             self.remove_orphan_tx(&orphan.tx.proposal_short_id()).await;
+                            self.broadcast_tx(Some(orphan.peer), orphan.tx.hash(), with_vm_2021);
+                            orphan_queue.push_back(orphan.tx);
                         }
-                        if reject.is_malformed_tx() {
-                            self.ban_malformed(orphan.peer, format!("reject {}", reject));
+                        Err(reject) => {
+                            if !is_missing_input(&reject) {
+                                self.remove_orphan_tx(&orphan.tx.proposal_short_id()).await;
+                            }
+                            if reject.is_malformed_tx() {
+                                self.ban_malformed(orphan.peer, format!("reject {}", reject));
+                            }
+                            break;
                         }
-                        break;
                     }
                 }
             }
@@ -525,8 +665,13 @@ impl TxPoolService {
             .any(|pt| snapshot.transaction_exists(&pt.tx_hash()))
     }
 
-    pub(crate) fn broadcast_tx(&self, origin: Option<PeerIndex>, tx_hash: Byte32) {
-        if let Err(e) = self.tx_relay_sender.send((origin, tx_hash)) {
+    pub(crate) fn broadcast_tx(
+        &self,
+        origin: Option<PeerIndex>,
+        tx_hash: Byte32,
+        with_vm_2021: bool,
+    ) {
+        if let Err(e) = self.tx_relay_sender.send((origin, with_vm_2021, tx_hash)) {
             error!("tx-pool broadcast_tx internal error {}", e);
         }
     }
@@ -556,21 +701,137 @@ impl TxPoolService {
         self.network.ban_peer(peer, DEFAULT_BAN_TIME, reason);
     }
 
+    pub(crate) async fn _resumeble_process_tx(
+        &self,
+        tx: TransactionView,
+        limit_cycles: Cycle,
+    ) -> Result<ProcessResult, Reject> {
+        let tx_hash = tx.hash();
+        // non contextual verify first
+        non_contextual_verify(&self.consensus, &tx)?;
+
+        if self.chunk_contains(&tx).await || self.orphan_contains(&tx).await {
+            return Err(Reject::Duplicated(tx.hash()));
+        }
+
+        let (tip_hash, snapshot, rtx, status, fee, tx_size) = self.pre_check(tx).await?;
+
+        let cached = self.fetch_tx_verify_cache(&tx_hash).await;
+        let tip_header = snapshot.tip_header();
+        let tx_env = status.with_env(tip_header);
+
+        let completed = if let Some(ref entry) = cached {
+            match entry {
+                CacheEntry::Completed(completed) => {
+                    TimeRelativeTransactionVerifier::new(
+                        &rtx,
+                        &self.consensus,
+                        snapshot.as_ref(),
+                        &tx_env,
+                    )
+                    .verify()
+                    .map_err(Reject::Verification)?;
+                    *completed
+                }
+                CacheEntry::Suspended(_) => {
+                    return Ok(ProcessResult::Suspended);
+                }
+            }
+        } else {
+            let consensus = snapshot.consensus();
+            let data_provider = snapshot.as_data_provider();
+            let is_chunk_full = !self.is_chunk_full().await;
+
+            let entry = block_in_place(|| {
+                let verifier =
+                    ContextualTransactionVerifier::new(&rtx, consensus, &data_provider, &tx_env);
+
+                let (ret, fee) = verifier
+                    .resumable_verify(limit_cycles)
+                    .map_err(Reject::Verification)?;
+
+                match ret {
+                    ScriptVerifyResult::Completed(cycles) => Ok(CacheEntry::completed(cycles, fee)),
+                    ScriptVerifyResult::Suspended(state) => {
+                        if is_chunk_full {
+                            let snap = Arc::new(state.try_into().map_err(Reject::Verification)?);
+                            Ok(CacheEntry::suspended(snap, fee))
+                        } else {
+                            Err(Reject::Full(
+                                "chunk".to_owned(),
+                                DEFAULT_MAX_CHUNK_TRANSACTIONS as u64,
+                            ))
+                        }
+                    }
+                }
+            })?;
+
+            match entry {
+                cached @ CacheEntry::Suspended(_) => {
+                    self.enqueue_suspended_tx(rtx.transaction.clone(), cached)
+                        .await?;
+                    return Ok(ProcessResult::Suspended);
+                }
+                CacheEntry::Completed(completed) => completed,
+            }
+        };
+
+        let entry = TxEntry::new(rtx, completed.cycles, fee, tx_size);
+
+        self.submit_entry(completed, tip_hash, entry, status)
+            .await?;
+
+        if cached.is_none() {
+            // update cache
+            let txs_verify_cache = Arc::clone(&self.txs_verify_cache);
+            tokio::spawn(async move {
+                let mut guard = txs_verify_cache.write().await;
+                guard.put(tx_hash, CacheEntry::Completed(completed));
+            });
+        }
+
+        Ok(ProcessResult::Completed(completed))
+    }
+
+    pub(crate) async fn is_chunk_full(&self) -> bool {
+        self.chunk.read().await.is_full()
+    }
+
+    pub(crate) async fn enqueue_suspended_tx(
+        &self,
+        tx: TransactionView,
+        cached: CacheEntry,
+    ) -> Result<(), Reject> {
+        let tx_hash = tx.hash();
+        let mut chunk = self.chunk.write().await;
+        if chunk.add_tx(tx) {
+            let mut guard = self.txs_verify_cache.write().await;
+            guard.put(tx_hash, cached);
+        }
+        Ok(())
+    }
+
     pub(crate) async fn _process_tx(
         &self,
         tx: TransactionView,
         max_cycles: Option<Cycle>,
-    ) -> Result<CacheEntry, Reject> {
+    ) -> Result<Completed, Reject> {
         let tx_hash = tx.hash();
 
         // non contextual verify first
         non_contextual_verify(&self.consensus, &tx)?;
 
+        if self.chunk_contains(&tx).await || self.orphan_contains(&tx).await {
+            return Err(Reject::Duplicated(tx.hash()));
+        }
+
         let (tip_hash, snapshot, rtx, status, fee, tx_size) = self.pre_check(tx).await?;
 
         let verify_cache = self.fetch_tx_verify_cache(&tx_hash).await;
-        let max_cycles = max_cycles.unwrap_or(self.tx_pool_config.max_tx_verify_cycles);
-        let verified = verify_rtx(&snapshot, &rtx, verify_cache, max_cycles)?;
+        let max_cycles = max_cycles.unwrap_or_else(|| self.consensus.max_block_cycles());
+        let tip_header = snapshot.tip_header();
+        let tx_env = status.with_env(tip_header);
+        let verified = verify_rtx(&snapshot, &rtx, &tx_env, &verify_cache, max_cycles)?;
 
         let entry = TxEntry::new(rtx, verified.cycles, fee, tx_size);
 
@@ -581,7 +842,7 @@ impl TxPoolService {
             let txs_verify_cache = Arc::clone(&self.txs_verify_cache);
             tokio::spawn(async move {
                 let mut guard = txs_verify_cache.write().await;
-                guard.put(tx_hash, verified);
+                guard.put(tx_hash, CacheEntry::Completed(verified));
             });
         }
 
@@ -597,6 +858,11 @@ impl TxPoolService {
     ) {
         let mut detached = LinkedHashSet::default();
         let mut attached = LinkedHashSet::default();
+        let hardfork_switch = snapshot.consensus().hardfork_switch();
+        let hardfork_during_detach =
+            check_if_hardfork_during_blocks(&hardfork_switch, &detached_blocks);
+        let hardfork_during_attach =
+            check_if_hardfork_during_blocks(&hardfork_switch, &attached_blocks);
 
         for blk in detached_blocks {
             detached.extend(blk.transactions().into_iter().skip(1))
@@ -610,22 +876,64 @@ impl TxPoolService {
         let fetched_cache = self.fetch_txs_verify_cache(retain.iter()).await;
 
         {
-            let mut tx_pool = self.tx_pool.write().await;
-            _update_tx_pool_for_reorg(
-                &mut tx_pool,
-                &attached,
-                detached_proposal_id,
-                snapshot,
-                &self.callbacks,
-            );
+            let txs_opt = {
+                // This closure is used to limit the lifetime of mutable tx_pool.
+                let mut tx_pool = self.tx_pool.write().await;
 
-            self.readd_dettached_tx(&mut tx_pool, retain, fetched_cache)
+                let txs_opt = if hardfork_during_detach || hardfork_during_attach {
+                    // The tx_pool is locked, remove all caches if has any hardfork.
+                    {
+                        self.txs_verify_cache.write().await.clear();
+                    }
+                    {
+                        self.chunk.write().await.clear();
+                    }
+
+                    Some(tx_pool.drain_all_transactions())
+                } else {
+                    None
+                };
+
+                _update_tx_pool_for_reorg(
+                    &mut tx_pool,
+                    &attached,
+                    detached_proposal_id,
+                    snapshot,
+                    &self.callbacks,
+                );
+                self.readd_dettached_tx(&mut tx_pool, retain, fetched_cache);
+
+                txs_opt
+            };
+
+            if let Some(txs) = txs_opt {
+                self.try_process_txs(txs).await;
+            }
         }
 
         {
             let mut orphan = self.orphan.write().await;
             orphan.remove_orphan_txs(attached.iter().map(|tx| tx.proposal_short_id()));
         }
+
+        {
+            let mut chunk = self.chunk.write().await;
+            chunk.remove_chunk_txs(attached.iter().map(|tx| tx.proposal_short_id()));
+        }
+
+        // update network fork switch each block
+        {
+            if !self.network.load_ckb2021() {
+                let epoch = self.snapshot().tip_header().epoch().number();
+                if self
+                    .consensus
+                    .hardfork_switch
+                    .is_vm_version_1_and_syscalls_2_enabled(epoch)
+                {
+                    self.network.init_ckb2021()
+                }
+            }
+        }
     }
 
     fn readd_dettached_tx(
@@ -641,8 +949,11 @@ impl TxPoolService {
             if let Ok((rtx, status)) = resolve_tx(tx_pool, tx_pool.snapshot(), tx) {
                 if let Ok(fee) = check_tx_fee(tx_pool, tx_pool.snapshot(), &rtx, tx_size) {
                     let verify_cache = fetched_cache.get(&tx_hash).cloned();
+                    let snapshot = tx_pool.snapshot();
+                    let tip_header = snapshot.tip_header();
+                    let tx_env = status.with_env(tip_header);
                     if let Ok(verified) =
-                        verify_rtx(tx_pool.snapshot(), &rtx, verify_cache, max_cycles)
+                        verify_rtx(snapshot, &rtx, &tx_env, &verify_cache, max_cycles)
                     {
                         let entry = TxEntry::new(rtx, verified.cycles, fee, tx_size);
                         if let Err(e) = _submit_entry(tx_pool, status, entry, &self.callbacks) {
@@ -654,12 +965,40 @@ impl TxPoolService {
         }
     }
 
+    // # Notice
+    //
+    // This method assumes that the inputs transactions are sorted.
+    async fn try_process_txs(&self, txs: Vec<TransactionView>) {
+        if txs.is_empty() {
+            return;
+        }
+        let total = txs.len();
+        let mut count = 0usize;
+        for tx in txs {
+            let tx_hash = tx.hash();
+            if let Err(err) = self._process_tx(tx, None).await {
+                error!("failed to process {:#x}, error: {:?}", tx_hash, err);
+                count += 1;
+            }
+        }
+        if count != 0 {
+            info!("{}/{} transactions are failed to process", count, total);
+        }
+    }
+
     pub(crate) async fn clear_pool(&mut self, new_snapshot: Arc<Snapshot>) {
         let mut tx_pool = self.tx_pool.write().await;
-        let config = tx_pool.config;
+        let config = tx_pool.config.clone();
         self.last_txs_updated_at = Arc::new(AtomicU64::new(0));
         *tx_pool = TxPool::new(config, new_snapshot, Arc::clone(&self.last_txs_updated_at));
     }
+
+    pub(crate) async fn save_pool(&mut self) {
+        let mut tx_pool = self.tx_pool.write().await;
+        if let Err(err) = tx_pool.save_into_file() {
+            error!("failed to save pool, error: {:?}", err)
+        }
+    }
 }
 
 type PreCheckedTx = (
@@ -679,36 +1018,63 @@ fn check_rtx(
     rtx: &ResolvedTransaction,
 ) -> Result<TxStatus, Reject> {
     let short_id = rtx.transaction.proposal_short_id();
+    let tip_header = snapshot.tip_header();
+    let proposal_window = snapshot.consensus().tx_proposal_window();
+    let hardfork_switch = snapshot.consensus().hardfork_switch();
     if snapshot.proposals().contains_proposed(&short_id) {
+        let resolve_opts = {
+            let tx_env = TxStatus::Proposed.with_env(tip_header);
+            let epoch_number = tx_env.epoch_number(proposal_window);
+            ResolveOptions::new().apply_current_features(hardfork_switch, epoch_number)
+        };
         tx_pool
-            .check_rtx_from_proposed(rtx)
+            .check_rtx_from_proposed(rtx, resolve_opts)
             .map(|_| TxStatus::Proposed)
     } else {
-        tx_pool.check_rtx_from_pending_and_proposed(rtx).map(|_| {
-            if snapshot.proposals().contains_gap(&short_id) {
-                TxStatus::Gap
-            } else {
-                TxStatus::Fresh
-            }
-        })
+        let tx_status = if snapshot.proposals().contains_gap(&short_id) {
+            TxStatus::Gap
+        } else {
+            TxStatus::Fresh
+        };
+        let resolve_opts = {
+            let tx_env = tx_status.with_env(tip_header);
+            let epoch_number = tx_env.epoch_number(proposal_window);
+            ResolveOptions::new().apply_current_features(hardfork_switch, epoch_number)
+        };
+        tx_pool
+            .check_rtx_from_pending_and_proposed(rtx, resolve_opts)
+            .map(|_| tx_status)
     }
 }
 
 fn resolve_tx(tx_pool: &TxPool, snapshot: &Snapshot, tx: TransactionView) -> ResolveResult {
     let short_id = tx.proposal_short_id();
+    let tip_header = snapshot.tip_header();
+    let proposal_window = snapshot.consensus().tx_proposal_window();
+    let hardfork_switch = snapshot.consensus().hardfork_switch();
     if snapshot.proposals().contains_proposed(&short_id) {
+        let resolve_opts = {
+            let tx_env = TxStatus::Proposed.with_env(tip_header);
+            let epoch_number = tx_env.epoch_number(proposal_window);
+            ResolveOptions::new().apply_current_features(hardfork_switch, epoch_number)
+        };
         tx_pool
-            .resolve_tx_from_proposed(tx)
+            .resolve_tx_from_proposed(tx, resolve_opts)
             .map(|rtx| (rtx, TxStatus::Proposed))
     } else {
-        tx_pool.resolve_tx_from_pending_and_proposed(tx).map(|rtx| {
-            let status = if snapshot.proposals().contains_gap(&short_id) {
-                TxStatus::Gap
-            } else {
-                TxStatus::Fresh
-            };
-            (rtx, status)
-        })
+        let tx_status = if snapshot.proposals().contains_gap(&short_id) {
+            TxStatus::Gap
+        } else {
+            TxStatus::Fresh
+        };
+        let resolve_opts = {
+            let tx_env = tx_status.with_env(tip_header);
+            let epoch_number = tx_env.epoch_number(proposal_window);
+            ResolveOptions::new().apply_current_features(hardfork_switch, epoch_number)
+        };
+        tx_pool
+            .resolve_tx_from_pending_and_proposed(tx, resolve_opts)
+            .map(|rtx| (rtx, tx_status))
     }
 }
 
@@ -718,18 +1084,28 @@ fn _submit_entry(
     entry: TxEntry,
     callbacks: &Callbacks,
 ) -> Result<(), Reject> {
+    let tx_hash = entry.transaction().hash();
     match status {
         TxStatus::Fresh => {
-            tx_pool.add_pending(entry.clone());
-            callbacks.call_pending(tx_pool, &entry);
+            if tx_pool.add_pending(entry.clone()) {
+                callbacks.call_pending(tx_pool, &entry);
+            } else {
+                return Err(Reject::Duplicated(tx_hash));
+            }
         }
         TxStatus::Gap => {
-            tx_pool.add_gap(entry.clone());
-            callbacks.call_pending(tx_pool, &entry);
+            if tx_pool.add_gap(entry.clone()) {
+                callbacks.call_pending(tx_pool, &entry);
+            } else {
+                return Err(Reject::Duplicated(tx_hash));
+            }
         }
         TxStatus::Proposed => {
-            tx_pool.add_proposed(entry.clone())?;
-            callbacks.call_proposed(tx_pool, &entry, true);
+            if tx_pool.add_proposed(entry.clone())? {
+                callbacks.call_proposed(tx_pool, &entry, true);
+            } else {
+                return Err(Reject::Duplicated(tx_hash));
+            }
         }
     }
     Ok(())
@@ -771,7 +1147,7 @@ fn _update_tx_pool_for_reorg(
         if snapshot.proposals().contains_proposed(entry.key()) {
             let tx_entry = entry.get();
             entries.push((
-                Some(CacheEntry::new(tx_entry.cycles, tx_entry.fee)),
+                Some(CacheEntry::completed(tx_entry.cycles, tx_entry.fee)),
                 tx_entry.clone(),
             ));
             entry.remove();
@@ -783,14 +1159,14 @@ fn _update_tx_pool_for_reorg(
         if snapshot.proposals().contains_proposed(entry.key()) {
             let tx_entry = entry.get();
             entries.push((
-                Some(CacheEntry::new(tx_entry.cycles, tx_entry.fee)),
+                Some(CacheEntry::completed(tx_entry.cycles, tx_entry.fee)),
                 tx_entry.clone(),
             ));
             entry.remove();
         } else if snapshot.proposals().contains_gap(entry.key()) {
             let tx_entry = entry.get();
             gaps.push((
-                Some(CacheEntry::new(tx_entry.cycles, tx_entry.fee)),
+                Some(CacheEntry::completed(tx_entry.cycles, tx_entry.fee)),
                 tx_entry.clone(),
             ));
             entry.remove();
@@ -816,3 +1192,31 @@ fn _update_tx_pool_for_reorg(
         }
     }
 }
+
+// # Notice
+//
+// This method assumes that the inputs blocks are sorted.
+fn check_if_hardfork_during_blocks(
+    hardfork_switch: &HardForkSwitch,
+    blocks: &VecDeque<BlockView>,
+) -> bool {
+    if blocks.is_empty() {
+        false
+    } else {
+        // This method assumes that the hardfork epochs are sorted and unique.
+        let hardfork_epochs = hardfork_switch.script_result_changed_at();
+        if hardfork_epochs.is_empty() {
+            false
+        } else {
+            let epoch_first = blocks.front().unwrap().epoch().number();
+            let epoch_next = blocks
+                .back()
+                .unwrap()
+                .epoch()
+                .minimum_epoch_number_after_n_blocks(1);
+            hardfork_epochs
+                .into_iter()
+                .any(|hardfork_epoch| epoch_first < hardfork_epoch && hardfork_epoch <= epoch_next)
+        }
+    }
+}
